import yargs from 'yargs';
import * as clack from '@clack/prompts';
import {CommandHandler} from './common';
import * as tty from 'tty';

export interface CommandArgs {
	aliases: string[];
	command: string;
	parameters: string[];
	describe: string;
	handler: CommandHandler;
	interactiveMode?: boolean;
}

type PromptValueArgs = [Record<string, string>, ...never[]]
	| [string, string]
	| [[string, string], ...never[]]
	| [[string, string][], ...never[]];

class DummyWriteStream extends tty.WriteStream {
	private bufferInternal = Buffer.from('');

	constructor() {
		super(0);
	}

	write(input: Uint8Array | string, _encoding?: BufferEncoding | ((err?: Error) => void), _cb?: (err?: Error) => void): boolean {
		this.bufferInternal = Buffer.concat([this.bufferInternal, Buffer.from(input)]);
		// noop
		return true;
	}

	get buffer(): Buffer {
		return this.bufferInternal;
	}
}

export interface TestModeResult {
	exitCode?: number;
	stdout: DummyWriteStream;
	stderr: DummyWriteStream;
}

export interface CliOptions {
	interactiveMode?: {
		option?: string;
		alias?: string;
		intro?: string;
		cancelled?: string;
		outro?: string;
		describe?: string;
	}
}

export class Cli {
	private testMode = false;

	private promptValues = {} as Record<string, unknown>;

	private testModeResult: TestModeResult = {
		exitCode: undefined as number | undefined,
		stdout: new DummyWriteStream(),
		stderr: new DummyWriteStream(),
	};

	private readonly cli = yargs()
		.help()
		.fail(false)
		.strict(false);

	constructor(private readonly options: CliOptions) {
		// noop
	}

	command({ parameters, command, interactiveMode = true, ...args }: CommandArgs): Cli {
		const thisCli = this;
		const { interactiveMode: theInteractiveMode = {} } = this.options;
		const {
			option = 'interactive',
			alias = 'i',
			intro = 'Please provide the following values:',
			cancelled = 'Operation cancelled.',
			outro = 'Thank you!',
			describe = 'Interactive mode',
		} = theInteractiveMode;

		const commandArgs = {
			...args,
			command: `${command} ${parameters.join(' ').replace(/</g, '[').replace(/>/g, ']')}`,
			usage: parameters.map((p) => `${command} ${p}`).join('\n'),
			handler: async function handler(commandArgs: Record<string, unknown>) {
				const self = this as any;

				const interactiveLong = option in commandArgs ? Boolean(commandArgs.interactive) : false;
				const interactiveShort = alias in commandArgs ? Boolean(commandArgs.i) : false;
				const interactive = interactiveLong || interactiveShort;
				let buildArgs = {
					...commandArgs,
				};

				const exit = (code: number) => {
					if (thisCli.testMode) {
						thisCli.testModeResult.exitCode = code;
						return;
					}
					process.exit(code);
				}

				if (interactive) {
					// TODO properly filter attributes
					const clackGroup = self.optional
						.filter((optional: { cmd: string[] }) => {
							const {
								'$0': _$0,
								i: _i,
								interactive: _interactive,
								test: _test,
								_: __,
								...rest
							} = buildArgs;

							return !Object.keys(rest).includes(optional.cmd[0]);
						})
						.reduce(
							(group: Record<string, string>, optional: { cmd: string[], describe: string }) => ({
								...group,
								[optional.cmd[0]]: optional.describe ?? optional.cmd[0],
							}),
							{} as Record<string, string>,
						);

					const prompt = async (messages: Record<string, string>) => {
						const messagesEntries = Object.entries(messages);

						if (messagesEntries.length > 0) {
							clack.intro(intro);
							const values = await clack.group(
								Object.fromEntries(
									messagesEntries.map(([key, value]) => [key, () => clack.text({message: value})])
								),
								{
									onCancel: () => {
										clack.cancel(cancelled);
										exit(1);
									}
								}
							);
							clack.outro(outro);
							return values;
						}

						return {};
					};

					const promptedArgs = thisCli.testMode ? thisCli.promptValues : await prompt(clackGroup);
					buildArgs = {
						...buildArgs,
						...promptedArgs,
					};
				}

				const returnCode = await args.handler({
					self,
					interactive,
					logger: {
						log: (message: unknown) => {
							const stdout = thisCli.testMode ? thisCli.testModeResult.stdout : process.stdout;
							stdout.write(`${message?.toString()}\n`);
						},
						error: (message: unknown) => {
							const stderr = thisCli.testMode ? thisCli.testModeResult.stderr : process.stderr;
							stderr.write(`${message?.toString()}\n`);
						},
					},
					send: exit,
					args: buildArgs,
				});

				exit(returnCode ? returnCode : 0);
			},
		} as any;

		if (interactiveMode) {
			commandArgs.options = {
				...(commandArgs.options ?? {}),
				[option]: {
					alias,
					describe,
					type: 'boolean',
					default: false,
					hidden: true,
				},
			};
		}

		this.cli.command(commandArgs);
		return this;
	}

	async run(args: string[]): Promise<void> {
		await this.cli.parse(args);
	}

	test() {
		this.testMode = true;
		this.promptValues = {};
		const testDelegate = {
			promptValue: (...args: PromptValueArgs) => {
				const [testArg] = args;

				if (typeof testArg === 'string') {
					const [key, value] = args as [string, unknown];
					this.promptValues[key] = this.promptValues[key] ?? value;
				}

				if (Array.isArray(testArg)) {
					const [key, value] = testArg as [string, unknown];
					this.promptValues[key] = this.promptValues[key] ?? value;
				}

				if (typeof testArg === 'object' && testArg !== null) {
					Object.entries(testArg).forEach(([key, value]) => {
						this.promptValues[key] = this.promptValues[key] ?? value;
					});
				}

				return testDelegate;
			},
			run: async (args: string[]): Promise<TestModeResult> => {
				await this.cli.parse(args);
				return this.testModeResult;
			},
		};
		return testDelegate;
	}
}

export const createCli = (options = {} as CliOptions) => {
	return new Cli(options);
};
